/*
 * Hanashizume no Matsuri for X
 * Copyright (c) 2000, 2001 TF <tf@denpa.org>
 * All rights reserverd.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *      This product includes software developed by TF.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: Panel.c,v 1.52 2001/03/07 16:01:48 tf Exp $
 */
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Vendor.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/time.h>

#include "PanelP.h"

#include "hana.h"

#define Disp       XtDisplay(pw)
#define Win        XtWindow(pw)
#define Screen     XtScreen(pw)
#define Root       XRootWindowOfScreen(XtScreen(pw))
#define Width      pw->core.width
#define Height     pw->core.height
#define PW         pw->panel

#define offset(field) XtOffsetOf(PanelRec, field)
static XtResource resources[] = {
  { XtNloadGame, XtCLoadGame, XtRBool, sizeof(Bool),
    offset(panel.load_game), XtRBool, NULL },
  { XtNredraw, XtCRedraw, XtRBool, sizeof(Bool),
    offset(panel.redraw), XtRBool, NULL },
  { XtNfont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
   offset(panel.font), XtRString, XtDefaultFont },
};
#undef offset

typedef struct {
  struct timeval start;
  int msec;
} timer;

static void Initialize();
static void Resize(Widget);
static Boolean SetValues();
static void Destroy(Widget);
static XtGeometryResult QueryGeometry();
static void Map(Widget, XEvent *, String *, Cardinal *);
static void Redisplay(Widget, XEvent *, Region);
static void ClickBtn1(Widget, XEvent *, String *, Cardinal *);
static void NewPage(PanelWidget);
static void DrawChar(PanelWidget, Drawable, int *, int *, int *, int, int, long, Bool);
static int DrawString(PanelWidget, Drawable, int *, long, int, Bool);
static void SetCG(PanelWidget, const char *, int, ImageAttribute *);
static void SetCG_FADE(PanelWidget, const char *, int, ImageAttribute *);
static void RedrawText(PanelWidget, Drawable);
// static void Enter(Widget, XtPointer, XEvent *, Boolean *);
static void Enter(Widget, XEvent *, String *, Cardinal *);
static void SetAlternative(PanelWidget);
static void Motion(Widget, XEvent *, String *, Cardinal *);
static void RedrawAlternative(PanelWidget, Drawable);
static void Skip(Widget, XEvent *, String *, Cardinal *);
static void CG_Mode(Widget, XEvent *, String *, Cardinal *);
static void SelectUp(Widget, XEvent *, String *, Cardinal *);
static void SelectDown(Widget, XEvent *, String *, Cardinal *);
static void LineFeed(PanelWidget);
static void Flash(PanelWidget, const char *);
static void ColorMask(PanelWidget, const char *);
static void Fadeout(PanelWidget);
static void Highlight(PanelWidget);
static void GameEngine(XtPointer, XtIntervalId *);
static void init_timer(timer *);
static void get_timer(timer *);

static XtActionsRec actions[] = {
  {"click_btn1", ClickBtn1},
  {"map_notify", Map},
  {"motion", Motion},
  {"key_tab", Skip},
  {"key_space", CG_Mode},
  {"enter_notify", Enter},
  {"select_up", SelectUp},
  {"select_down", SelectDown}
};

static char translations[] = 
"<Motion>: motion()\n"
"<MapNotify>: map_notify()\n"
"<EnterNotify>: enter_notify()\n"
"<Key>space: key_space()\n"
"<Key>Tab: key_tab()\n"
"Alt<Key>s: key_tab()\n"
"<Key>Return: click_btn1()\n"
"<Key>KP_Enter: click_btn1()\n"
"<Btn1Down>: click_btn1()\n"
"<Key>Up: select_up()\n"
"<Key>KP_Up: select_up()\n"
"<Key>k: select_up()\n"
"<Key>Down: select_down()\n"
"<Key>KP_Down: select_down()\n"
"<Key>j: select_down()";

PanelClassRec panelClassRec = {
  {
  /* superclass             */ (WidgetClass) &compositeClassRec,
  /* class_name             */ "Panel",
  /* widget_size            */ sizeof(PanelRec),
  /* class_initialize       */ NULL,
  /* class_part_initialize  */ NULL,
  /* class_inited           */ FALSE,
  /* initialize             */ Initialize,
  /* initialize_hook        */ NULL,
  /* realize                */ XtInheritRealize,
  /* actions                */ actions,
  /* num_actions            */ XtNumber(actions),
  /* resources              */ resources,
  /* num_resourece          */ XtNumber(resources),
  /* xrm_class              */ NULLQUARK,
  /* compress_motion        */ TRUE,
  /* compress_exposure      */ TRUE,
  /* compress_enterleave    */ TRUE,
  /* visible_interest       */ FALSE,
  /* destroy                */ Destroy,
  /* resize                 */ Resize,
  /* expose                 */ Redisplay,
  /* set_values             */ SetValues,
  /* set_values_hook        */ NULL,
  /* set_values_almoset     */ XtInheritSetValuesAlmost,
  /* get_values_hook        */ NULL,
  /* accept_focus           */ NULL,
  /* version                */ XtVersion,
  /* callback_private       */ NULL,
  /* tm_table               */ translations,
  /* query_geometry         */ QueryGeometry,
  /* display_accelator      */ XtInheritDisplayAccelerator,
  /* extention              */ NULL
  },{
  XtInheritGeometryManager,
  XtInheritChangeManaged,
  XtInheritInsertChild,
  XtInheritDeleteChild,
  NULL
  },{
  /* foo                    */ 0
  }
};
WidgetClass panelWidgetClass = (WidgetClass)&panelClassRec;

static long
GetPixelByName(Widget core, const char *name)
{
  XColor scr, exact;
  XAllocNamedColor(XtDisplay(core), core->core.colormap, 
		   name, &scr, &exact);
  return exact.pixel;
}

static void
Initialize(Widget wreq, Widget wnew)
{
  PanelWidget pw = (PanelWidget) wnew;

  PW.load_game = False;
  
  Width = 640;
  Height = 480;
  PW.cur_x = 0;
  PW.cur_y = 0;
  PW.selection = -1;

  open_archives();
  make_u2e_table();
  PW.gc = DefaultGC(Disp, XScreenNumberOfScreen(Screen));

  /* font */
  InitFont(Disp, PW.gc, PW.font);

  /* color */
  PW.black_pixel = GetPixelByName(wnew, "black");
  PW.white_pixel = GetPixelByName(wnew, "white");
  PW.gray_pixel = GetPixelByName(wnew, "gray");
  PW.yellow_pixel = GetPixelByName(wnew, "yellow");
  PW.prevgray = False;

  /* text cache */
  PW.text_cache_len = 0;

  /* scenario */
  init_scenario();
  
  /* music */
  PW.music_playing = False;
  music_init();

  PW.timewait_id = None;
}

static Boolean
SetValues(Widget wcur, Widget wreq, Widget wnew)
{
  PanelWidget cur = (PanelWidget) wcur;
  PanelWidget new = (PanelWidget) wnew;
  XtAppContext app = XtWidgetToApplicationContext((Widget)new);
  
  Boolean redraw = False;

  return redraw;
}

/* don't change this value, because it depends on a_map[] table in image.c */
#define FADE_STEP 16
/* the unit is msec */
#define FADE_TIME 1500

static int
DrawString
(PanelWidget pw, Drawable dst, int *buf, long color, int attr, Bool cache)
{
  int x, y, x0;

  if (PW.cur_y > YMAX) {
    NewPage(pw);
  }
  if (scn_stat.newpage == True) {
    NewPage(pw);
  }

  x0 = x = XOFFSET + PW.cur_x * finfo.width;
  y = YOFFSET + PW.cur_y * finfo.pitch;

  while (*buf != 0) {
    if (PW.cur_x + 1> XMAX) {
      PW.cur_x = 0;
      PW.cur_y++;
      x = 0;
      y += finfo.pitch;
      if (PW.cur_y > YMAX) {
	fprintf(stderr, "text is too long.\n");
	NewPage(pw);
      }
    }
    DrawChar(pw, dst, &x, &y, buf, 1, attr, color, cache);
    PW.cur_x++;
    buf++;
  }

  return x - x0;
}

static void
LineFeed(PanelWidget pw)
{
  PW.text_cache[PW.text_cache_len++] = 0x0a;

  PW.cur_x = 0;
  if (++PW.cur_y > YMAX) {
    fprintf(stderr, "text is too long.\n");
  }
}

static void
NewPage(PanelWidget pw)
{
  PW.cur_x = 0;
  PW.cur_y = 0;

  /* clear text cache */
  PW.text_cache_len = 0;

  /* store the status of top */
  strcpy(scn_stat.pe_script, scn_stat.script);
  scn_stat.pe_block = scn_stat.block;
  scn_stat.pe_offset = scn_stat.offset;
  scn_stat.pe_alternative = scn_stat.alternative;
  
  /* clear screen */
  if (XtIsRealized((Widget)pw)) {
    XCopyArea(Disp, PW.bgimage->pixmap, Win, PW.bgimage->gc, 0, 0, Width, Height, 0, 0);
    XSync(Disp, False);
  }
  
  scn_stat.newpage = False;
}

static void
RedrawText(PanelWidget pw, Drawable dst)
{
  int i, x, y, cx, cy;
  
  cx = XOFFSET;
  cy = YOFFSET;
  for (i = 0, x = 0, y = 0; i < PW.text_cache_len; i++) {
    if (PW.text_cache[i] == 0x0a) {
      x = 0;
      cx = XOFFSET;
      y++;
      cy += finfo.pitch;
      continue;
    }
    if (x + 1> XMAX) {
      x = 0;
      cx = XOFFSET;
      y++;
      cy += finfo.pitch;
    }
    DrawChar(pw, dst, &cx, &cy, &PW.text_cache[i], 1, CH_SHADOW, PW.white_pixel, False);
    x++;
  }

  if (PW.state == STATE_ALTN) {
    RedrawAlternative(pw, Win);
  }
}

static void
DrawChar(PanelWidget pw, Drawable dst, int *x, int *y, int *code, int length, int attr, long color, Bool cache)
{
  if (cache) {
    PW.text_cache[PW.text_cache_len++] = *code;
  }

  PutChar(dst, code, x, y, color, PW.black_pixel, attr);
}

static void
Destroy(Widget w)
{
  PanelWidget pw = (PanelWidget) w;

  ImageDestroy(PW.bgimage);
  ImageDestroy(PW.workimage);
  ImageDestroy(PW.white);

  music_term();

  close_archives();
}

static void
Resize(Widget w)
{
  return;
}

static void
Map(Widget w, XEvent *ev, String *p, Cardinal *n)
{
  PanelWidget pw = (PanelWidget) w;

  PW.bgimage = ImageCreate(Disp, Win, Width, Height);
  PW.workimage = ImageCreate(Disp, Win, Width, Height);
  PW.white = ImageCreate(Disp, Win, Width, Height);
  SetWhite(PW.white, Width, Height);

  SetCG(pw, "CGBLACK", sys_stat.brightness, NULL);
    
  StartEngine(pw);
}

static void
Redisplay(Widget w, XEvent *ev, Region region)
{
  PanelWidget pw = (PanelWidget) w;

  if (XtIsRealized(w)) {
    XCopyArea(Disp, PW.bgimage->pixmap, Win, PW.bgimage->gc, 0, 0, Width, Height, 0, 0);
    RedrawText(pw, Win);
    XSync(Disp, False);
  }
}

void
ResetDisplay(PanelWidget pw)
{
  GRAYSCALE g = {False, 256, 256, 256};
  int density;
  
  if (scn_stat.state == STAT_OPENING ||
      scn_stat.state == STAT_ENDING ||
      scn_stat.alt_opening) {
    density = MAX_DENSITY;
  } else {
    density = sys_stat.brightness;
  }
  set_ximage(PW.bgimage, Width, Height, density, g);
  ImagePut(PW.bgimage);
  XCopyArea(Disp, PW.bgimage->pixmap, Win, PW.bgimage->gc, 0, 0, Width, Height, 0, 0);
  RedrawText(pw, Win);
}

static XtGeometryResult
QueryGeometry(Widget w, XtWidgetGeometry *proposed, XtWidgetGeometry *answer)
{
  PanelWidget pw = (PanelWidget) w;
  
  answer->request_mode = CWWidth | CWHeight;
  answer->width  = Width;
  answer->height = Height;
  
  return XtGeometryAlmost;
}

static void
GameEngine(XtPointer cl, XtIntervalId * id)
{
  PanelWidget pw = (PanelWidget) cl;
  unsigned long interval = 1;
  int density;

  XtAppContext app = XtWidgetToApplicationContext((Widget)pw);
  PW.command.type = NONE;
  ImageAttributeInit(&PW.command.attr);
  if (scn_stat.state == STAT_OPENING ||
      scn_stat.state == STAT_ENDING) {
    density = MAX_DENSITY;
  } else {
    density = sys_stat.brightness;
  }

  do_scenario(msg, &PW.command);

  if (scn_stat.alternative == True) {
    SetAlternative(pw);
    PW.command.type = ALTERNATIVE;
    sys_stat.skip = False;
  }

  switch (PW.command.type) {
  case NONE:
    /* fprintf(stderr, "No command is specified.\n"); */
    break;
  case PRINT_TEXT:
    DrawString(pw, Win, PW.command.ubuf, PW.white_pixel, CH_SHADOW, True);
    break;
  case LINE_FEED:
    LineFeed(pw);
    break;
  case NEW_PAGE:
    if (scn_stat.state == STAT_OPENING ||
	scn_stat.state == STAT_ENDING) {
      NewPage(pw);
      break;
    } else if (sys_stat.speed == SPEED_VFAST && !sys_stat.skip) {
      PW.state = STATE_WAIT;
    }
    scn_stat.newpage = True;
    break;
  case WAIT_CLICK:
    if (sys_stat.speed == SPEED_VFAST ||
	scn_stat.state == STAT_OPENING ||
	scn_stat.state == STAT_ENDING ||
	sys_stat.skip == True) {
      break;
    }
    PW.state = STATE_WAIT;
    break;
  case WAIT:
    if ((sys_stat.speed == SPEED_VFAST && scn_stat.state == STAT_PROC) ||
	sys_stat.skip == True) {
      interval = 0;
    } else {
      interval = PW.command.arg;
    }
    break;
  case SET_CG:
    // fprintf(stderr, "set CG[%s], arg %d\n", PW.command.buf, PW.command.arg);
    SetCG(pw, PW.command.buf, density, &PW.command.attr);
    RedrawText(pw, Win);
    break;
  case SET_CG_FADE:
    // fprintf(stderr, "set CG[%s], fade %d\n", PW.command.buf, PW.command.arg);
    if ((sys_stat.speed == SPEED_VFAST && scn_stat.state == STAT_PROC) ||
	sys_stat.skip == True || sys_stat.cg_fade == False) {
      SetCG(pw, PW.command.buf, density, &PW.command.attr);
    } else {
      SetCG_FADE(pw, PW.command.buf, density, &PW.command.attr);
    }
    RedrawText(pw, Win);
    break;
  case SET_CG_OMAKE:
    SetCG(pw, "CGHANA", density, &PW.command.attr);
    break;
  case ALTERNATIVE:
    PW.state = STATE_ALTN;
    break;
  case PLAY_MUSIC:
    if (PW.command.arg != 0) {
      if (PW.music_playing) {
	music_stop();
      }
      music_play(PW.command.arg, True);
      PW.music_playing = True;
    } else {
      if (PW.music_playing) {
	music_stop();
	PW.music_playing = False;
      }
    }
    break;
  case FLASH:
    Flash(pw, PW.command.buf);
    break;
  case COLOR_MASK:
    ColorMask(pw, PW.command.buf);
    break;
  case HIGHLIGHT:
    Highlight(pw);
    break;
  case FADE_OUT:
    Fadeout(pw);
    break;
  }

  if (PW.state != STATE_WAIT && PW.state != STATE_ALTN) {
    PW.timewait_id = XtAppAddTimeOut(app, interval, GameEngine, pw);
  }
}

static void
SetCG(PanelWidget pw, const char *name, int density, ImageAttribute *attr)
{
  RESINFO *r;
  GRAYSCALE grayscale = {False, 256, 256, 256};

  if (scn_stat.newpage) {
    NewPage(pw);
  }

  if (attr != NULL) {
    grayscale = attr->grayscale;
  }
  
  r = search_rlist(cg.rlist, name);
  if (r == NULL) {
    fprintf(stderr, "Can't find the image: %s.\n", name);
    exit(1);
  }
  if (PW.bgimage->rgbmap != NULL) {
    free(PW.bgimage->rgbmap);
  }
  strncpy(scn_stat.cgname, name, NAMELEN);
  PW.bgimage->rgbmap = get_jpeg(cg.fd, r->offset, Width, Height);
  set_ximage(PW.bgimage, Width, Height, density, grayscale);
  ImagePut(PW.bgimage);
  
  if (XtIsRealized((Widget)pw)) {
    XCopyArea(Disp, PW.bgimage->pixmap, Win, PW.bgimage->gc, 0, 0, Width, Height, 0, 0);
    XSync(Disp, False);
  }
  
  if (attr != NULL) {
    PW.prevgray = grayscale.enable;
  }

  scn_stat.image = PW.bgimage;	/* specify an image to be saved */
}

static void
SetCG_FADE(PanelWidget pw, const char *name, int density, ImageAttribute *attr)
{
  RESINFO *r;
  unsigned char *tmp;
  int i, th;
  GRAYSCALE grayscale = {False, 256, 256, 256};
  timer t;

  if (scn_stat.newpage) {
    NewPage(pw);
  }

  if (!XtIsRealized((Widget)pw)) {
    return;
  }
  if (attr != NULL) {
    grayscale = attr->grayscale;
    if (PW.prevgray) {
      grayscale.enable = True;
    }
  }
  
  r = search_rlist(cg.rlist, name);
  if (r == NULL) {
    fprintf(stderr, "Can't find the image: %s.\n", name);
    exit(1);
  }
  if (PW.workimage->rgbmap != NULL) {
    free(PW.workimage->rgbmap);
  }
  PW.workimage->rgbmap = get_jpeg(cg.fd, r->offset, Width, Height);

  init_timer(&t);
  for (i = 0, th = 0; i < FADE_STEP; i ++, th += FADE_TIME / FADE_STEP) {
    get_timer(&t);
    if (t.msec > th) {
#ifdef DEBUG
      fprintf(stderr, "skip: %d > %d\n", t.msec, th);
#endif
      continue;
    }
    mix_ximages(PW.bgimage, PW.workimage, Width, Height, i, density, grayscale);
    ImagePut(PW.bgimage);
    XCopyArea(Disp, PW.bgimage->pixmap, Win, PW.bgimage->gc, 0, 0, Width, Height, 0, 0);
    XSync(Disp, False);
  }
  tmp = PW.bgimage->rgbmap;
  PW.bgimage->rgbmap = PW.workimage->rgbmap;
  PW.workimage->rgbmap = tmp;
  strncpy(scn_stat.cgname, name, NAMELEN);
  set_ximage(PW.bgimage, Width, Height, density, grayscale);
  ImagePut(PW.bgimage);

  XCopyArea(Disp, PW.bgimage->pixmap, Win, PW.bgimage->gc, 0, 0, Width, Height, 0, 0);
  XSync(Disp, False);

  if (attr != NULL) {
    PW.prevgray = attr->grayscale.enable;
  }

  scn_stat.image = PW.bgimage;	/* XXX for save image */
}

void
StartEngine(PanelWidget pw)
{
  if (PW.timewait_id == None) {
    XtAppContext app = XtWidgetToApplicationContext((Widget)pw);
    PW.timewait_id = XtAppAddTimeOut(app, 0, GameEngine, pw);
  }
}

void
StopEngine(PanelWidget pw)
{
  if (PW.timewait_id) {
    XtRemoveTimeOut(PW.timewait_id);
    PW.timewait_id = None;
  }
}

static void
ClickBtn1(Widget w, XEvent *ev, String *p, Cardinal *n)
{
  PanelWidget pw = (PanelWidget) w;

  sys_stat.skip = False;

  if (PW.state == STATE_WAIT) {
    PW.timewait_id = None;
    PW.state = STATE_PROC;
    StartEngine(pw);
  } else if (PW.state == STATE_ALTN) {
    if (PW.selection < 0) { /* no selection */
      return;
    }
    scn_stat.alt_val = PW.selection;
    PW.timewait_id = None;
    PW.state = STATE_PROC;
    scn_stat.alt_opening = False;
    StartEngine(pw);
  }
}

static void
Enter(Widget pw, XEvent *ev, String *p, Cardinal *n)
{
  XSetInputFocus(Disp, Win, RevertToNone, CurrentTime);
}

static void
SetAlternative(PanelWidget pw)
{
  int i, j;
  int tabbing, esc;
  int maxlen = 0;
  static int whitespace[] = {0x3000, 0};
  PW.selection = -1;

  for (i = 0; i < scn_stat.alt_num; i++) {
    fprintf(stderr, "%d:", i + 1);
    puteuc_string(scn_stat.alt[i]);
    fprintf(stderr, "\n");
    for (j = 0; ; j++) {
      if (scn_stat.alt[i][j] == 0) {
	break;
      }
    }
    if (j > maxlen) {
      maxlen = j;
    }
  }

  if (scn_stat.alt_opening) {
    NewPage(pw);
    for (i = 0; i < MARGIN_Y_OPENING; i++) {
      LineFeed(pw);
    }
  } else {
    if (PW.cur_x != 0) {
      LineFeed(pw);
    }
    LineFeed(pw);
  }
  
  for (i = 0; i < scn_stat.alt_num; i++) {
    if (scn_stat.alt_opening) {
      tabbing = MARGIN_X_OPENING;
    } else {
      tabbing = (XMAX - maxlen) / 2;
    }
    for (j = 0; j < tabbing; j++) {
      DrawString(pw, Win, whitespace, PW.gray_pixel, CH_BORDER, False);
    }
    PW.alt_cur_x[i] = PW.cur_x;
    PW.alt_cur_y[i] = PW.cur_y;
    PW.alt_x[i][0] = XOFFSET + PW.cur_x * finfo.width;
    PW.alt_y[i][0] = YOFFSET + PW.cur_y * finfo.pitch;
    esc = DrawString(pw, Win, scn_stat.alt[i], PW.gray_pixel, CH_BORDER, False);
    PW.alt_x[i][1] = PW.alt_x[i][0] + esc;
    PW.alt_y[i][1] = PW.alt_y[i][0] + finfo.pitch - YPAD;
    LineFeed(pw);
  }
}

static void
Motion(Widget w, XEvent *ev, String *p, Cardinal *n)
{
  PanelWidget pw = (PanelWidget) w;

  XMotionEvent *me = (XMotionEvent *)ev;
  int i;
  int prev_selection = PW.selection;
  
  if (PW.state != STATE_ALTN) {
    return;
  }

  for (i = 0; i < scn_stat.alt_num; i++) {
    if (me->x > PW.alt_x[i][0] &&
	me->y > PW.alt_y[i][0] &&
	me->x < PW.alt_x[i][1] &&
	me->y < PW.alt_y[i][1]) {
      PW.selection = i;
      break;
    }
    PW.selection = -1;
  }
  if (PW.selection != prev_selection) {
    RedrawAlternative(pw, Win);
  }
}

static void
RedrawAlternative(PanelWidget pw, Drawable dst)
{
  long color;
  int i;

  for (i = 0; i < scn_stat.alt_num; i++) {
    PW.cur_x = PW.alt_cur_x[i];
    PW.cur_y = PW.alt_cur_y[i];
    if (i == PW.selection) {
      color = PW.yellow_pixel;
    } else {
      color = PW.gray_pixel;
    }
    DrawString(pw, Win, scn_stat.alt[i], color, CH_BORDER, False);
  }
}

static void
Skip(Widget w, XEvent *ev, String *p, Cardinal *n)
{
  PanelWidget pw = (PanelWidget) w;

  sys_stat.skip = True;

  if (PW.state == STATE_WAIT) {
    PW.timewait_id = None;
    PW.state = STATE_PROC;
    StartEngine(pw);
  }
}

static void
CG_Mode(Widget w, XEvent *ev, String *p, Cardinal *n)
{
  PanelWidget pw = (PanelWidget) w;
  static int prev_stat;
  GRAYSCALE g = {False, 256, 256, 256};
  int density;

  g.enable = PW.prevgray;

  if (PW.state != STATE_CG) {
    set_ximage(PW.bgimage, Width, Height, 32, g);
    ImagePut(PW.bgimage);
    XCopyArea(Disp, PW.bgimage->pixmap, Win, PW.bgimage->gc, 0, 0, Width, Height, 0, 0);
    XSync(Disp, False);
  
    prev_stat = PW.state;
    PW.state = STATE_CG;
  } else {
    ResetDisplay(pw);
    if (prev_stat == STATE_ALTN) {
      RedrawAlternative(pw, Win);
    }
    XSync(Disp, False);
    
    PW.timewait_id = None;
    PW.state = prev_stat;
    if (PW.state != STATE_ALTN &&
	PW.state != STATE_WAIT) {
      StartEngine(pw);
    }
  }
}

static void
SelectUp(Widget w, XEvent *ev, String *p, Cardinal *n)
{
  PanelWidget pw = (PanelWidget) w;

  if (PW.state != STATE_ALTN) {
    return;
  }

  if (PW.selection < 0) {
    PW.selection = 0;
  } else {
    if (--PW.selection < 0) {
      PW.selection = scn_stat.alt_num - 1;
    }
  }
  RedrawAlternative(pw, Win);
}

static void
SelectDown(Widget w, XEvent *ev, String *p, Cardinal *n)
{
  PanelWidget pw = (PanelWidget) w;

  if (PW.state != STATE_ALTN) {
    return;
  }

  if (PW.selection < 0) {
    PW.selection = scn_stat.alt_num - 1;
  } else {
    if (++PW.selection > scn_stat.alt_num - 1) {
      PW.selection = 0;
    }
  }
  RedrawAlternative(pw, Win);
}

static void
Flash(PanelWidget pw, const char *buf)
{
  int msec;
  const char *p = buf;
  
  if (!XtIsRealized((Widget)pw)) {
    return;
  }
  
  while (*p != '\0') {
    if (isdigit(*p)) {
      sscanf(p, "%d", &msec);
      do {
	p++;
      } while (isdigit(*p));
      XCopyArea(Disp, PW.white->pixmap, Win, PW.white->gc, 0, 0, Width, Height, 0, 0);
      usleep(msec);
      XSync(Disp, False);
      XCopyArea(Disp, PW.bgimage->pixmap, Win, PW.bgimage->gc, 0, 0, Width, Height, 0, 0);
      usleep(msec);
      XSync(Disp, False);
      continue;
    } else if (*p == 'i') {
      p++;
      continue;
    } else {
      fprintf(stderr, "unknown token\n");
      break;
    }
  }
}

static void
Highlight(PanelWidget pw)
{
  SetWhite(PW.bgimage, Width, Height);
  ImagePut(PW.bgimage);
  XCopyArea(Disp, PW.bgimage->pixmap, Win, PW.white->gc, 0, 0, Width, Height, 0, 0);
  RedrawText(pw, Win);
  if (PW.state == STATE_ALTN) {
    RedrawAlternative(pw, Win);
  }
  XSync(Disp, False);
}

static void
Fadeout(PanelWidget pw)
{
  if ((sys_stat.speed == SPEED_VFAST && scn_stat.state == STAT_PROC) ||
      sys_stat.skip == True) {
    return;
  }
    if ((sys_stat.speed == SPEED_VFAST && scn_stat.state == STAT_PROC) ||
	sys_stat.skip == True || sys_stat.cg_fade == False) {
      SetCG(pw, "CGBLACK", sys_stat.brightness, NULL);
    } else {
      SetCG_FADE(pw, "CGBLACK", sys_stat.brightness, NULL);
    }
}

static void
ColorMask(PanelWidget pw, const char *buf)
{
  GRAYSCALE g = {True, 256, 256, 256};
  int i = 0;
  char tmp[12];
  const char *p = buf;

  while (isdigit(*p) || *p == ',') {
    tmp[i++] = *p++;
  }
  tmp[i] = '\0';

  /* currently, other options aren't processed */
  
  sscanf(tmp, "%d,%d,%d", &g.r, &g.g, &g.b);
  
  set_ximage(PW.bgimage, Width, Height, sys_stat.brightness, g);
  ImagePut(PW.bgimage);
  XCopyArea(Disp, PW.bgimage->pixmap, Win, PW.bgimage->gc, 0, 0, Width, Height, 0, 0);
  RedrawText(pw, Win);
  if (PW.state == STATE_ALTN) {
    RedrawAlternative(pw, Win);
  }
  XSync(Disp, False);
}

void
LoadGame(PanelWidget pw)
{
  sys_stat.skip = False;
  
  /* clear text */
  NewPage(pw);
  
  /* set CG */
  if (strlen(scn_stat.cgname) > 0) {
    if (scn_stat.state == STAT_OPENING) {
      SetCG(pw, scn_stat.cgname, MAX_DENSITY, NULL);
    } else {
      SetCG(pw, scn_stat.cgname, sys_stat.brightness, NULL);
    }
  }
  /* effect should be restored, too */

  /* set music */
  if (scn_stat.track) {
    music_play(scn_stat.track, True);
  } else {
    music_stop();
  }

  /* currently, we can restart only from top of the block */
  if (scn_stat.state == STAT_OPENING) {
    scn_stat.init_state = STAT_OPENING;
    scn_stat.state = STAT_INIT;
  } else if (scn_stat.alternative) {
    scn_stat.alternative = False;
    scn_stat.state = STAT_ALTN;
  } else {
    scn_stat.init_state = STAT_PROC;
    scn_stat.state = STAT_INIT;
  }

  if (PW.state == STATE_WAIT || PW.state == STATE_ALTN) {
    PW.timewait_id = None;
    PW.state = STATE_PROC;
  }
  
  StartEngine(pw);
}

static void
init_timer(timer *tp)
{
  gettimeofday(&tp->start, 0);
  tp->msec = 0;
}

static void
get_timer(timer *tp)
{
  struct timeval tv;
  long s, us;

  gettimeofday(&tv, 0);
  s = tv.tv_sec - tp->start.tv_sec;
  us = tv.tv_usec - tp->start.tv_usec;

  tp->msec = s * 1000 + us / 1000;
}
