/*
 * Hanashizume no Matsuri for X
 * Copyright (c) 2000, 2001 TF <tf@denpa.org>
 * All rights reserverd.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *      This product includes software developed by TF.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: menu.c,v 1.12 2001/01/08 14:13:37 tf Exp $
 */
#include <stdio.h>
#include <time.h>

#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>

#include <X11/IntrinsicP.h>
#include <X11/CoreP.h>
#include <X11/ShellP.h>

#include "SmeCascade.h"
#include "CascadeMenu.h"
#include "Panel.h"

#include "bitmap/select.xbm"
#include "bitmap/SmeCascade.xbm"

#include "hana.h"

static int bmp_selected;
static int bmp_cascaded;
static Bool menu_button_up = False;

Widget menuShell;
Widget speedmenuShell;
Widget musicmenuShell;
Widget brightnessmenuShell;

Widget saveItem;
Widget loadItem;
Widget screenItem;
Widget captureItem;
Widget titleItem;
Widget textItem;
Widget fadeItem;
Widget speedItem;
Widget brightnessItem;
Widget nextItem;
Widget musicItem;
Widget quitItem;

FileBrowser loadbrowser, savebrowser, capturebrowser;

Widget speed[5];
char speedName[5][7] = {
  "Read", "Vfast", "Fast", "Normal", "Slow"
};
int speedValue[5] = {
  -1, SPEED_VFAST, SPEED_FAST, SPEED_NORMAL, SPEED_SLOW
};

Widget brightness[4];
char brightnessName[4][9] = {
  "bright", "normal", "dark", "verydark",
};
int brightnessValue[4] = {
  BG_DENSITY1, BG_DENSITY2, BG_DENSITY3, BG_DENSITY4,
};

Widget music[NMUSIC];
char musicName[NMUSIC][5] = {
  "None", "Cdda", "Midi", "Mp3"
};
int musicValue[NMUSIC] = {
  MUSIC_NONE, MUSIC_CDDA, MUSIC_MIDI, MUSIC_MP3
};

Widget none, cdda, midi;

static Widget *menuItem[] = {
  &saveItem,
  &loadItem,
  &screenItem,
  &captureItem,
  &titleItem,
  &textItem,
  &fadeItem,
  &speedItem,
  &brightnessItem,
  &nextItem,
  &musicItem,
  &quitItem,
};

static Widget create_speed_menu(void);
static Widget create_brightness_menu(void);
static Widget create_music_menu(void);

void
MenuPopup(Widget w, XEvent *e, String *p, Cardinal *n)
{
  Dimension width, height;
  ShellWidget sw = (ShellWidget)menuShell;

  if (sw->shell.popped_up) {
    XtPopdown(menuShell);
    return;
  }
  
  if (w != menuShell) {
    XtCallActionProc(w, "XawPositionSimpleMenu", e, p, *n);
#if 0
    XtPopup(menuShell, XtGrabNone);
#else
    XtPopup(menuShell, XtGrabNonexclusive);
    XtGrabPointer(menuShell, True,
		  ButtonPressMask|ButtonReleaseMask|ButtonMotionMask,
		  GrabModeAsync, GrabModeAsync, None, None, CurrentTime);
#endif
    menu_button_up = False;
  } else {
    XtVaGetValues(menuShell, XtNwidth, &width, XtNheight, &height, NULL);
    if (e->xbutton.x < 0 || e->xbutton.x > width
	|| e->xbutton.y < 0 || e->xbutton.y > height) {
      XtPopdown(menuShell);
    }
  }
}

static void
MenuSelect(Widget w, XEvent *e, String *p, Cardinal *n)
{
  unsigned int i;
  Position x, y;
  Dimension width, height;
  
  if (w != menuShell) {
    return;
  }

  if (!menu_button_up) {
    menu_button_up = True;
    return;
  }

  for(i = 0; i < XtNumber(menuItem); i++) {
    XtVaGetValues(*menuItem[i],
		  XtNx, &x,
		  XtNy, &y,
		  XtNwidth, &width,
		  XtNheight, &height,
		  NULL );
    if (e->xbutton.x >= x && e->xbutton.x <= (x + width)
	&& e->xbutton.y >= y && e->xbutton.y <= (y + height)) {
      XtCallActionProc(w, "highlight", e, p, *n);
      XtCallCallbacks(*menuItem[i], XtNcallback, NULL);
      XtCallActionProc(w, "unhighlight", e, p, *n);
      XtPopdown(menuShell);
      return;
    }
  }
}

static void
MenuHighlight(Widget w, XEvent *e, String *p, Cardinal *n)
{
  unsigned int i;
  Position x, y;
  Dimension width, height;
  
  for(i = 0; i < XtNumber(menuItem); i++ ) {
    menu_button_up = True;
    XtVaGetValues(*menuItem[i],
		  XtNx, &x,
		  XtNy, &y,
		  XtNwidth, &width,
		  XtNheight, &height,
		  NULL );
    if (e->xbutton.x > x && e->xbutton.x < (x + width)
	&& e->xbutton.y > y && e->xbutton.y < (y + height)) {
      XtCallActionProc(w, "highlight", e, p, *n);
      return;
    }
  }
}

static void
MenuSave(Widget w, XEvent *e, String *p, Cardinal *n)
{
#define DATE_FILE_LEN 17
  char datefile[DATE_FILE_LEN];
  
  time_t s;
  struct tm *t;
  s = time(NULL);
  t = localtime(&s);
  snprintf(datefile, DATE_FILE_LEN, "%04d-%02d-%02d_%02d:%02d", 
	   t->tm_year + 1900, 
	   t->tm_mon + 1, 
	   t->tm_mday, 
	   t->tm_hour, 
	   t->tm_min);

  setdir(&savebrowser, path.savedir);
  set_filename(&savebrowser, datefile);
  XtPopup(savebrowser.shell, XtGrabNonexclusive);
  XtSetKeyboardFocus(savebrowser.shell, savebrowser.name);
}

static void
MenuLoad(Widget w, XEvent *e, String *p, Cardinal *n)
{
  setdir(&loadbrowser, path.savedir);
  XtPopup(loadbrowser.shell, XtGrabNonexclusive);
  XtSetKeyboardFocus(loadbrowser.shell, loadbrowser.name);
}

static void
MenuScreen(Widget w, XEvent *e, String *p, Cardinal *n)
{
  fprintf(stderr, "full screen mode(not implemented)\n");
}

static void
MenuCapture(Widget w, XEvent *e, String *p, Cardinal *n)
{
  char savename[NAMELEN];

  setdir(&capturebrowser, path.imagedir);
  strncpy(savename, scn_stat.cgname, NAMELEN);
  strncat(savename, ".png", NAMELEN);
  set_filename(&capturebrowser, savename);
  XtPopup(capturebrowser.shell, XtGrabNonexclusive);
  XtSetKeyboardFocus(capturebrowser.shell, capturebrowser.name);
}

static void
MenuTitle(Widget w, XEvent *e, String *p, Cardinal *n)
{
  StopEngine((PanelWidget)core);

  free_script();
  init_scenario();

  strcpy(scn_stat.script, "0x1d");
  scn_stat.block = 0;
  scn_stat.offset = TITLE_OFFSET;

  scn_stat.alt_opening = True;
  scn_stat.state = STAT_OPENING;

  LoadGame((PanelWidget)core);
}

static void
MenuText(Widget w, XEvent *e, String *p, Cardinal *n)
{
  fprintf(stderr, "show/hide text(not implemented)\n");
}

static void
MenuFade(Widget w, XEvent *e, String *p, Cardinal *n)
{
  if (sys_stat.cg_fade) {
    sys_stat.cg_fade = False;
    XtVaSetValues(fadeItem, XtNleftBitmap, None, NULL);
  } else {
    sys_stat.cg_fade = True;
    XtVaSetValues(fadeItem, XtNleftBitmap, bmp_selected, NULL);
  }
}

static void
MenuNext(Widget w, XEvent *e, String *p, Cardinal *n)
{
  fprintf(stderr, "next\n");
}

static void
MenuQuit(Widget w, XEvent *e, String *p, Cardinal *n)
{
  quit_proc(top, NULL, NULL);
}

Widget 
AddMenuCascade(Widget menu, const char *name, Dimension lm, Dimension rm, Widget cascade)
{
  Widget entry;

  entry = XtVaCreateManagedWidget(name, smeCascadeObjectClass, menu,
				  XtNmenu, cascade,
				  XtNleftMargin, lm, XtNrightMargin, rm,
				  XtNrightBitmap, bmp_cascaded,
				  NULL);
  return entry;
}

Widget
AddMenu(Widget menu, const char *name, XtCallbackProc callback,
	Dimension lm, Dimension rm, Pixmap pixmenu)
{
  Widget w;
  
  w = XtVaCreateManagedWidget(name, smeBSBObjectClass, menu, 
			      XtNleftMargin, lm,
			      XtNrightMargin, rm,
			      XtNleftBitmap, pixmenu,
			      NULL);
  XtAddCallback(w, XtNcallback, callback, NULL);

  return w;
}

static XtActionsRec menuActions[] = {
  { "MenuPopup", (XtActionProc)MenuPopup },
  { "MenuSelect", (XtActionProc)MenuSelect },
  { "MenuHighlight", (XtActionProc)MenuHighlight },
};

void CreateMenu(XtAppContext app, Widget top)
{
  bmp_cascaded = XCreateBitmapFromData(XtDisplay(top), 
	XtWindow(top), cascade_bits, cascade_width, cascade_height);

  bmp_selected = XCreateBitmapFromData(XtDisplay(top), 
	XtWindow(top), select_bits, select_width, select_height);
  
  XtAppAddActions(app, menuActions, XtNumber(menuActions));

  menuShell = XtVaCreatePopupShell("menu", cascadeMenuWidgetClass, top, NULL);

  CreateBrowser(menuShell, &savebrowser, "savebrowser", SAVE_BROWSER);
  saveItem = AddMenu(menuShell, "Save", (XtCallbackProc)MenuSave,
		     18, 18, None);
  CreateBrowser(menuShell, &loadbrowser, "loadbrowser", LOAD_BROWSER);
  loadItem = AddMenu(menuShell, "Load", (XtCallbackProc)MenuLoad,
		     18, 18, None);
  screenItem = AddMenu(menuShell, "Screen", (XtCallbackProc)MenuScreen,
		       18, 18, None);
  CreateBrowser(menuShell, &capturebrowser, "capturebrowser", CAPTURE_BROWSER);
  captureItem = AddMenu(menuShell, "Capture", (XtCallbackProc)MenuCapture,
			18, 18, None);
  titleItem = AddMenu(menuShell, "Title", (XtCallbackProc)MenuTitle,
		      18, 18, None);
  textItem = AddMenu(menuShell, "Text", (XtCallbackProc)MenuText,
		     18, 18, bmp_selected);
  if (sys_stat.cg_fade) {
    fadeItem = AddMenu(menuShell, "Fade", (XtCallbackProc)MenuFade,
		       18, 18, bmp_selected);
  } else {
    fadeItem = AddMenu(menuShell, "Fade", (XtCallbackProc)MenuFade,
		       18, 18, None);
  }

  speedmenuShell = create_speed_menu();
  speedItem = AddMenuCascade(menuShell, "Speed", 18, 18, speedmenuShell);

  brightnessmenuShell = create_brightness_menu();
  brightnessItem = AddMenuCascade(menuShell, "Brightness", 18, 18, brightnessmenuShell);

  nextItem = AddMenu(menuShell, "Next", (XtCallbackProc)MenuNext,
		     18, 18, None);

  musicmenuShell = create_music_menu();
  musicItem = AddMenuCascade(menuShell, "Music", 18, 18, musicmenuShell);

  quitItem = AddMenu(menuShell, "Quit", (XtCallbackProc)MenuQuit,
		     18, 18, None);
}

/*
 * Text display settings
 */
static void
MenuSpeed(Widget w, XEvent *e, String *p, Cardinal *n)
{
  int i;

  if (w == speed[0]) {
    if (sys_stat.skip) {
      sys_stat.skip = False;
      XtVaSetValues(speed[0], XtNleftBitmap, None, NULL);
    } else {
      sys_stat.skip = True;
      XtVaSetValues(speed[0], XtNleftBitmap, bmp_selected, NULL);
    }
    return;
  }

  for (i = 1; i < sizeof(speed) / sizeof(speed[0]); i++) {
    if (w == speed[i]) {
      sys_stat.speed = speedValue[i];
      XtVaSetValues(speed[i], XtNleftBitmap, bmp_selected, NULL);
    } else {
      XtVaSetValues(speed[i], XtNleftBitmap, None, NULL);
    }
  }
}

static Widget create_speed_menu()
{
  Widget menu;
  int i;

  menu = XtVaCreatePopupShell("speed", cascadeMenuWidgetClass, top, NULL);

  speed[0] = AddMenu(menu, speedName[0],
		     (XtCallbackProc)MenuSpeed, 18, 18, None);

  for (i = 1; i < sizeof(speed) / sizeof(speed[0]); i++) {
    if (sys_stat.speed == speedValue[i]) {
      speed[i] = AddMenu(menu, speedName[i],
			 (XtCallbackProc)MenuSpeed, 18, 18, bmp_selected);
    } else {
      speed[i] = AddMenu(menu, speedName[i],
			 (XtCallbackProc)MenuSpeed, 18, 18, None);
    }
  }

  return menu;
}

/*
 * Brightness of background image on text display mode
 */
static void
MenuBrightness(Widget w, XEvent *e, String *p, Cardinal *n)
{
  int i;

  for (i = 0; i < sizeof(brightness) / sizeof(brightness[0]); i++) {
    if (w == brightness[i]) {
      sys_stat.brightness = brightnessValue[i];
      XtVaSetValues(brightness[i], XtNleftBitmap, bmp_selected, NULL);
    } else {
      XtVaSetValues(brightness[i], XtNleftBitmap, None, NULL);
    }
  }
  ResetDisplay((PanelWidget)core);
}

static Widget create_brightness_menu()
{
  Widget menu;
  int i;

  menu = XtVaCreatePopupShell("brightness", cascadeMenuWidgetClass, top, NULL);

  for (i = 0; i < sizeof(brightness) / sizeof(brightness[0]); i++) {
    if (sys_stat.brightness == brightnessValue[i]) {
      brightness[i] = AddMenu(menu, brightnessName[i],
			 (XtCallbackProc)MenuBrightness, 18, 18, bmp_selected);
    } else {
      brightness[i] = AddMenu(menu, brightnessName[i],
                         (XtCallbackProc)MenuBrightness, 18, 18, None);
    }
  }
  
  return menu;
}

/*
 * Music settings
 */
static void
MenuMusic(Widget w, XEvent *e, String *p, Cardinal *n)
{
  int i, tmp_track;

  for (i = 0; i < sizeof(music) / sizeof(music[0]); i++) {
    if (w == music[i]) {
      fprintf(stderr, "music: %s\n", musicName[i]);
      XtVaSetValues(music[i], XtNleftBitmap, bmp_selected, NULL);
      if (sys_stat.music_mode != musicValue[i]) {
	tmp_track = scn_stat.track;
	music_term();
	sys_stat.music_mode = musicValue[i];
	music_init();
	scn_stat.track = tmp_track;
	if (scn_stat.track) {
	  music_play(scn_stat.track, True);
	}
      }
    } else {
      XtVaSetValues(music[i], XtNleftBitmap, None, NULL);
    }
  }
}

static Widget create_music_menu()
{
  Widget menu;
  int i;

  menu = XtVaCreatePopupShell("music", cascadeMenuWidgetClass, top, NULL);

  for (i = 0; i < sizeof(music) / sizeof(music[0]); i++) {
    if (sys_stat.music_mode == musicValue[i]) {
      music[i] = AddMenu(menu, musicName[i],
			 (XtCallbackProc)MenuMusic, 18, 18, bmp_selected);
    } else {
      music[i] = AddMenu(menu, musicName[i],
			 (XtCallbackProc)MenuMusic, 18, 18, None);
    }
  }

  return menu;
}

void load_dialog()
{
  MenuLoad((Widget)NULL, NULL, NULL, NULL);
}
